import json

from django.http import HttpResponse, HttpRequest
from django.views.decorators.csrf import csrf_exempt
from django.shortcuts import get_object_or_404

from vpos.models import Transaction

@csrf_exempt
def watch_transaction_confirmation(request: HttpRequest, transaction_id: str) -> HttpResponse:
    """View to watch vPOS API Webhook with Transaction Confirmation"""
    if request.method == 'POST':
        # TODO: (security) check api signature (vPOS dependency)
        # ...
        transaction: Transaction = get_object_or_404(Transaction,
            id=transaction_id)
        data: dict = json.loads(request.body)
        if not transaction.payment and data.get('id') == transaction.key:
            if transaction.confirm(data):
                return HttpResponse(status=200)
            return HttpResponse(status=409)
        return HttpResponse(status=403)
    return HttpResponse(status=405)



from vpos.models import Transaction


def create(mobile: str,
        amount: str,
        parent: Transaction = None) -> Transaction:
    """
    Creates new Payment or Refund Transaction
    If parent is set creates a Refund Transaction
    """
    if not parent:
        return Transaction.objects.create_payment(
            mobile=mobile, amount=amount)
    return Transaction.objects.create_refund(
        parent=parent)
    


# Generated by Django 3.2.7 on 2021-11-23 22:35

from django.db import migrations, models
import django.db.models.deletion
import uuid
import vpos.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, verbose_name='id')),
                ('key', models.CharField(default=None, editable=False, max_length=100, null=True, verbose_name='location id')),
                ('amount', models.DecimalField(decimal_places=2, editable=False, max_digits=12, verbose_name='amount')),
                ('mobile', models.CharField(editable=False, max_length=15, validators=[vpos.validators.PhoneAOValidator()], verbose_name='mobile number')),
                ('type', models.CharField(choices=[('payment', 'Payment'), ('refund', 'Refund')], editable=False, max_length=7, verbose_name='type')),
                ('requested', models.BooleanField(default=False, editable=False, verbose_name='was requested')),
                ('data', models.JSONField(default=dict, editable=False, verbose_name='vpos transaction data')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='updated at')),
                ('parent', models.OneToOneField(blank=True, default=None, editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='refund', to='vpos.transaction', verbose_name='parent transaction')),
            ],
            options={
                'verbose_name': 'Transaction',
                'verbose_name_plural': 'Transactions',
            },
        ),
    ]



from django.conf import settings
from django.utils.translation import gettext_lazy as _

from vpos.exceptions import VposConfigurationError as Err


DEFAULTS: dict = {
    # required
    'POS_ID': None,
    'TOKEN': None,
    'URL': None,
    'MODE': 'production',
    'VPOS_SUPERVISOR_CARD': None,
    'VPOS_BASE_URL': 'https://vpos.ao/api/v1',
    # optionals
    'VPOS_TEST_SUPERVISOR_CARD': '9610123456123412341234123456789012345'}


VPOS_STATUS_REASON: dict = {
    # client
    '3000': _('Refused by client'),
    # Processor
    '2010': _('Request was refused by the processor'),
    '2009': _('Parent transaction is too old to be refunded'),
    '2008': _('Invalid merchant email'),
    '2007': _('Invalid or Inactive supervisor card'),
    '2006': _('Insufficient funds in POS available for refund'),
    '2005': _('POS is closed and unable to accept transactions'),
    '2004': _('Request timed-out and was refused by the processor'),
    '2003': _('Card or network daily limit exceeded'),
    '2002': _('Refused by the card issuer'),
    '2001': _("Insufficient funds in client's account"),
    '2000': _('Generic processor error'),
    # Gateway
    '1003': _('Parent transaction ID of refund request is not an accepted Payment'),
    '1002': _('Gateway is not authorized to execute transactions on the specified POS'),
    '1001': _('Request timed-out and will not be processed'),
    '1000': _('Generic gateway error')}


class VposSettings:

    __defaults: dict

    def __init__(self, defaults: dict = None, user_settings: dict = None) -> None:
        if user_settings:
            self.__settings = user_settings
        self.__defaults = defaults or DEFAULTS

    @property
    def settigs(self) -> dict:
        if not hasattr(self, '__settings'):
            return getattr(settings, 'VPOS', {})
        return self.__settings
    
    def validate(self):
        """configurations validation"""
        for attr in dir(self):
            if 'validate_' in attr:
                getattr(self, attr)()
    
    def validate_mode(self):
        modes = ('production', 'sandbox')
        if self.MODE not in modes:
            raise Err(
                "Invalid MODE: '%s'. Must be " 
                "production or sandbox" % self.MODE)
    
    def validate_token(self):
        if not self.POS_ID:
            raise Err('TOKEN is required')
    
    def validate_url(self):
        if not self.POS_ID:
            raise Err('URL is required for callback transaction confirmation')
    
    def validate_vpos_base_url(self):
        if not self.POS_ID:
            raise Err('VPOS_BASE_URL is required')
    
    def validate_vpos_supervisor_card(self):
        if not self.POS_ID:
            raise Err('VPOS_SUPERVISOR_CARD is required')
    
    def validate_pos_id(self):
        if not self.POS_ID:
            raise Err('POS_ID is required')

    def get_supervisor_card(self) -> str:
        if self.MODE == 'sandbox':
            return self.VPOS_TEST_SUPERVISOR_CARD
        return self.VPOS_SUPERVISOR_CARD

    def __getattr__(self, attr: str):
        if attr not in self.__defaults:
            raise AttributeError("Invalid VPOS setting: '%s'" % attr)
        try:
            value = self.settigs[attr]
        except KeyError:
            value = self.__defaults[attr]
        setattr(self, attr, value)
        return value
    

conf = VposSettings()




from django.apps import AppConfig


class VposConfig(AppConfig):

    name = 'vpos'
    verbose_name = 'Django vPOS'

    def ready(self) -> None:
        from vpos.configs import conf
        conf.validate()



import time
import requests

from typing import Union
from requests import Response
from vpos.configs import conf


class VposAPI:

    """
    Vpos API Interaction
    Class for vpos.models.Transaction
    """

    __idempotency_key: str

    def __init__(self, idempotency_key: str) -> None:
        self.__idempotency_key = idempotency_key
    
    def create(self, **kwargs) -> Union[None, str]:
        """
        Create new Payment or Refund
        Expect this named args
            - type (str)
            - mobile (str|None)
            - amount (str|None)
            - parent_id (str|None)
            - polling: (bool)
        Returns location string
        """
        r = self.post('/transactions',
            data=self.__get_data_for_new_transaction(**kwargs))
        if r.status_code == 202:
            return r.headers.get('location')
        return None

    def check(self, request_id: str, wait: bool = False) -> Union[dict, None]:
        """
        Check Transaction Queued/Running Status
        Returns True if already completd else retuns False
        f payment process is running (waiting user answer, in queue or else)
        Will get the eta and wait (if True) some time to try the request again.
        When finished, will return the transaction data (by following the location header)
        """
        
        r = self.get(f'/requests/{request_id}')

        if r.status_code == 200:
            data: dict = r.json()
            if (eta := data.get('eta')) and wait:
                try:
                    seconds = int(float(eta)) + 1
                    time.sleep(seconds if seconds <= 90 else 90)
                except:
                    return None
                return self.check(request_id=request_id, wait=False)
            elif eta is None:
                return data # transaction data       
        return None

    # ---------------------------------------------------------------------
    # Base API Calls With Headers Configured

    def get(self, path: str, **kwargs) -> Response:
        url: str = f'{self.base_url}{path}'
        with requests.get(url, headers=self.__headers, **kwargs) as r:
            return r
        
    def post(self, path: str, data: dict = {}, **kwargs) -> Response:
        url: str = f'{self.base_url}{path}'
        with requests.post(url, json=data,
                headers=self.__headers, **kwargs) as r:
            return r
    
    def put(self, path: str, data: dict = {}, **kwargs) -> Response:
        url: str = f'{self.base_url}{path}'
        with requests.put(url, json=data,
                headers=self.__headers, **kwargs) as r:
            return r
    
    def delete(self, path: str, data: dict = {}, **kwargs) -> Response:
        url: str = f'{self.base_url}{path}'
        with requests.delete(url, json=data,
                headers=self.__headers, **kwargs) as r:
            return r

    @property
    def __headers(self) -> dict:
        headers = {
            'Content-Type': 'application/json',
            'Idempotency-Key': self.__idempotency_key,
            'Authorization': f'Bearer {conf.TOKEN}'}
        return headers
    
    @property
    def base_url(self) -> str:
        return conf.VPOS_BASE_URL
    
    @property
    def callback_url(self) -> str:
        return f'{conf.URL}/{self.__idempotency_key}'
    
    @property
    def vpos_id(self) -> str:
        return conf.VPOS_ID
    
    def __get_data_for_new_transaction(self, **kwargs) -> dict:
        """
        Expect this named args
            - type (str)
            - mobile (str|None)
            - amount (str|None)
            - parent_id (str|None)
            - polling: (bool)
        """
        if kwargs['type'] == 'refund':
            data = {
                'type': 'refund',
                'parent_transaction_id': kwargs.get('parent_id'),
                'supervisor_card': conf.get_supervisor_card()}
        else:
            data = {
                'type': 'payment',
                'pos_id': conf.POS_ID,
                'mobile': kwargs.get('mobile'),
                'amount': kwargs.get('amount')}

        if not kwargs.get('polling'):
            data['callback_url'] = self.callback_url
        return data



import re
from django.core import validators
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _


@deconstructible
class PhoneAOValidator(validators.RegexValidator):
    
    regex = r'^(?:(\+244|00244))?(9)(1|2|3|4|9)([\d]{7,7})$'
    default_replace = r'\2\3\4'
    message =  _('Invalid national phone number of angola')

    @classmethod
    def match(cls, string):
        return re.match(cls.regex, string)
    
    @classmethod
    def clean_number(cls, phone: str):
        return re.sub(cls.regex, cls.default_replace, phone)


# default
PhoneValidator = PhoneAOValidator



from django.urls import path
from vpos.views import watch_transaction_confirmation


app_name = 'vpos'

urlpatterns: list = [
    path('<uuid:transaction_id>',
        watch_transaction_confirmation, name='confirmation'),
]



import uuid
from typing import Union

from django.db import models
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError

from vpos.configs import (
    VPOS_STATUS_REASON,
    conf)
from vpos.validators import PhoneValidator
from vpos.signals import transaction_completed
from vpos.api import VposAPI


class TransactionType(models.TextChoices):
    PAYMENT = 'payment', _('Payment')
    REFUND = 'refund', ('Refund')


class Manager(models.Manager):
    
    def create_refund(self, parent):
        """Creates a new Refund Transaction"""
        transaction = self.model(
            parent=parent,
            type=TransactionType.REFUND,
            mobile=parent.mobile,
            amount=parent.amount)
        if conf.MODE == 'production':
            transaction.full_clean()
            if parent.rejected or parent.is_refund:
                raise ValidationError(
                    _("'parent' must be an accepted payment transaction"))
        transaction.save()
        return transaction
    
    def create_payment(self, mobile: str, amount: str):
        """Creates a new Refund Transaction"""
        transaction = self.model(
            type=TransactionType.PAYMENT,
            mobile=mobile,
            amount=amount,
            parent=None)
        if conf.MODE == 'production':
            transaction.full_clean()
        transaction.save()
        return transaction


class Transaction(models.Model):

    class Meta:
        verbose_name = _('Transaction')
        verbose_name_plural = _('Transactions')
    
    Type = TransactionType
    objects = Manager()
    api: VposAPI

    id = models.UUIDField(_('id'),
        primary_key=True, default=uuid.uuid4, editable=False)
    key = models.CharField(_('location id'),
        max_length=100, null=True, default=None, editable=False)
    amount = models.DecimalField(_('amount'),
        max_digits=12, decimal_places=2, editable=False)
    mobile = models.CharField(_('mobile number'),
        max_length=15, validators=[PhoneValidator()], editable=False)
    type = models.CharField(_('type'),
        max_length=7, choices=Type.choices, editable=False)
    requested = models.BooleanField(_('was requested'), default=False, editable=False)
    data = models.JSONField(_('vpos transaction data'), default=dict, editable=False)
    parent = models.OneToOneField(
        'self',
        on_delete=models.CASCADE,
        verbose_name=(_('parent transaction')),
        related_name='refund',
        blank=True,
        null=True,
        default=None,
        editable=False)
    
    created_at = models.DateTimeField(_('created at'), auto_now_add=True, db_index=True)
    updated_at = models.DateTimeField(_('updated at'), auto_now=True)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.api = VposAPI(
            idempotency_key=self.idempotency_key)

    @property
    def payment(self) -> Union[dict, None]:
        """Transaction Data from vPOS"""
        return self.data.get('transaction')
    
    @property
    def status(self) -> Union[str, None]:
        """Transaction Status from vPOS (accepted or rejected)"""
        if (payment := self.payment):
            return payment.get('status')
        return None
    
    @property
    def status_code(self) -> Union[int, None]:
        """vPOS Transaction Status Reason Code"""
        if (payment := self.payment):
            return payment.get('status_reason')
        return None
    
    @property
    def status_reason(self) -> Union[str, None]:
        """vPOS Transaction Status Reason Text"""
        if (code := self.status_code):
            return VPOS_STATUS_REASON.get(str(code))
        return None
    
    @property
    def accepted(self) -> bool:
        return self.status == 'accepted'
    
    @property
    def rejected(self) -> bool:
        return self.status == 'rejected'
    
    @property
    def is_payment(self) -> bool:
        return self.type == self.Type.PAYMENT
    
    @property
    def is_refund(self) -> bool:
        return self.type == self.Type.REFUND

    @property
    def idempotency_key(self) -> str:
        return str(self.id)
    
    @property
    def location(self) -> str:
        """location string provided by vPOS api header"""
        return self.data.get('location', '')

    def confirm(self, transaction_data: dict) -> bool:
        """Confirms transaction"""
        if not self.payment:
            self.__set_transaction_data(
                data=transaction_data)
            return True
        return False
    
    def check_payment(self, wait: bool = False) -> Union[dict, None]:
        """
        Checks transaction status from vPOS API,
        return transaction data if transaction accepted/rejected
        else returns None
        """
        if not self.payment:
            if (transaction := self.api.check(self.key, wait=wait)):
                self.__set_transaction_data(
                    data=transaction)
                return transaction
            return None
        return self.payment

    def request(self, polling: bool = False) -> bool:
        """Request Payment or Refund"""
        if not self.requested:
            if self.type == self.Type.REFUND:
                location = self.api.create(
                    type=str(self.Type.REFUND),
                    polling=polling,
                    parent_id=self.parent.key)
            else:
                location = self.api.create(
                    type=str(self.Type.PAYMENT),
                    mobile=PhoneValidator.clean_number(self.mobile),
                    amount=str(self.amount),
                    polling=polling)
            assert location is not None
            self.__set_key(location)
            return True
        return False
    
    def __set_key(self, location: str) -> None:
        if not self.key:
            self.key = location.split('/')[-1]
            self.data.update({'location': location})
            self.requested = True
            self.save()
    
    def __set_transaction_data(self, data: dict) -> None:
        if not self.payment:
            self.data.update({'transaction': data})
            self.save()
            self.__dispatch_transaction_completed()
    
    def __dispatch_transaction_completed(self):
        transaction_completed.send(
            self.__class__,
            transaction=self)
    